# -*- coding: utf-8 -*-
#
# This file is part of pytest-invenio.
# Copyright (C) 2017-2018 CERN.
#
# pytest-invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest fixtures for Invenio.

.. _quick-start:

Quick start
-----------
1. Define a module-scoped fixture named ``create_app`` that returns an
   application factory for your Invenio installation. If you are using
   Invenio-App, it's as simple as:

   .. code-block:: python

       # conftest.py
       from invenio_app.factory import create_ui

       @pytest.fixture(scope='module')
       def create_app():
           return create_ui


2. Write tests:

   .. code-block:: python

       # test_something.py

       def test_e2e(live_server, browser):
           browser.get(url_for('index', _external=True))

       def test_testclient(client):
           res = client.get('/api/')
           res.json == {'test-client': 'with-json-decoder'}

       def test_db(base_app, db):
           # Database with rollback

       def test_cli(cli_runner):
           result = cli_runner(mycmd)
           assert result.exit_code == 0

       def test_mailbox(appctx, mailbox):
           # ...
           assert len(mailbox) == 1

Running tests
-------------
Running tests with py.test is pretty simple. Your package might support the
standard way of running tests:

.. code-block:: console

    $ python setup.py test

Alternatively you can use the ``pytest`` command to run all or specific test
cases:

.. code-block:: console

    $ pytest
    $ pytest tests/test_something.py
    $ pytest tests/test_something.py::test_acase

Fixtures
--------
All available fixtures are documented in the API documentation
(see :ref:`fixtures`).

In addition to the ones provided by pytest-invenio, there are further fixtures
defined pytest-flask (see
`documentation <http://pytest-flask.readthedocs.io/en/latest/>`_ for details).

Structuring tests
-----------------
The pytest fixtures in pytest-invenio all work on *one* Flask application,
however most Invenio instances usually consits of *two* Flask applications:
UI and API. Thus, to use the pytest-invenio fixtures it's important to
understand how to structure your tests, and know exactly which application you
are dealing with.

Scope
~~~~~
Most of pytest-invenio fixtures are either *module* scoped or *function*
scoped.

* *Module* scoped fixtures are created/destroyed once per Python test file.
* *Function* scoped fixtures are created/destroyed per test.

The fixtures which creates the database and applications are module scoped,
hence, all tests in a Python file run against either the UI or the API
application, but not both.

.. note::

    All tests in a single file, run against the one and only one application
    (e.g. UI or REST).

Thus, in a single test file you cannot mix both UI and API tests, which is
normally not an issue.

Overriding fixtures
~~~~~~~~~~~~~~~~~~~
Pytest provides rich support for overriding fixtures at various level and
combined with module/function-scoped we can easily override fixtures. Also,
you can use ``conftest.py`` to define per-directory fixtures.

Following is an example of how fixtures overriding works:

.. code-block:: python

    # conftest.py:
    @pytest.fixture()
    def myfix():
        return 'root'

    # test_root.py
    def test_a(myfix):
        print(myfix)
        # will output "root"

    # a/conftest.py
    @pytest.fixture()
    def myfix(myfix):
        return myfix + '-a'

    # a/test_subdir.py
    def test_a(myfix):
        print(myfix)
        # will output "root-a"

Notice that:

* **Overriding:** In ``a/test_subdir.py`` the fixture ``myfix`` is coming from
  ``a/conftest.py`` which is overriding the fixture from ``conftest.py``. In
  ``test_root.py`` it's however the ``myfix`` fixture from ``conftest.py``
  being used.
* **Parent fixture:** In ``a/conftest.py``, the fixture ``myfix`` has access to
  the parent fixture from ``conftest.py``.

Recommend layout
~~~~~~~~~~~~~~~~
If you are using Invenio-App (recommended), then the following layout is
recommended:

.. code-block:: python

    # ### tests/conftest.py ###
    # Common application configuration goes here
    @pytest.fixture(scope='module')
    def app_config(app_config):
        app_config['MYCONF'] = True
        return app_config

    # ### tests/ui/conftest.py ###
    # UI tests goes in tests/ui/ folder.
    from invenio_app.factory import create_ui

    @pytest.fixture(scope='module')
    def create_app():
        return create_ui

    # ### tests/api/confest.py ###
    # API tests goes in tests/api/ folder.
    from invenio_app.factory import create_api

    @pytest.fixture(scope='module')
    def create_app():
        return create_api

    # ### tests/e2e/conftest.py ###
    # E2E tests (requring both UI/API) goes in tests/e2e/ folder.
    from invenio_app.factory import create_app as create_ui_api

    @pytest.fixture(scope='module')
    def create_app():
        return create_ui_api

Using above layout you essentially split your tests into three folders::

    tests/ui/
    tests/api/
    tests/e2e/

Each subfolder holds tests related to a specific application (UI or API).
The ``e2e`` folder holds tests that need both UI and API application (which
is typically the case for end-to-end tests). The E2E tests works by creating
both the UI and API applications and using a special WSGI middleware to
dispatch requests between both applications. Having two applications at the
same time, can however cause quite a lot of confusion so it is only recommended
for E2E tests.

Note, also in above example how all three applications are sharing the same
:py:data:`~fixtures.app_config` fixture.

.. note::

    You shouldn't feel bound to above structure. If you site grows large,
    you'll likely split tests into further subfolders. The important message
    from the recommended layout, is that you need **one folder per
    application**.

Application fixtures
--------------------
The package provides three different application fixtures:

* :py:data:`~fixtures.base_app`: Basic application fixture which creates the
  Flask application.
* :py:data:`~fixtures.appctx`: Same as the basic application fixture, but
  pushes an application context onto the stack (i.e. makes ``current_app``
  work).
* :py:data:`~fixtures.app`: Same as the basic application, but in addition it
  initializes the database and Elasticsearch indices.

All three fixtures depend on the same user-provided (i.e. you must define it)
fixture named ``create_app`` which must return an application factory (see
:ref:`quick-start`).

Customizing configuration
~~~~~~~~~~~~~~~~~~~~~~~~~
The application fixtures rely on fixtures such as
:py:data:`~fixtures.instance_path`, :py:data:`~fixtures.app_config`,
:py:data:`~fixtures.celery_config_ext`, :py:data:`~fixtures.db_uri`,
:py:data:`~fixtures.broker_uri` to inject configuration into the
application.

You can overwrite each of these fixtures at many different levels:

* **Global**: Override one or more of these fixtures in your global
  ``conftest.py`` to inject the same configuration in all applications.
* **Per-directory**: Override fixtures for a specific subdirectory by putting a
  ``conftest.py`` in the directory.
* **Per-file**: Fixtures can also be overwritten in specific modules. For
  instance you may want to customize the celery configuration only for a
  specific Python test file.

.. _views-testing:

Views testing
-------------
Views can easily be testing using the Flask test clients. Two test clients are
provided for convenience: ``base_client`` and ``client``. The only difference
is which application fixture they depend on:

.. code-block:: python

    def test_view1(base_client):
        # Depends on 'base_app' fixture
        base_client.get(url_for(..))

    def test_view2(client):
        # Depends on 'app' fixture
        client.get(url_for(..))

JSON responses
~~~~~~~~~~~~~~
The default Flask test client does not have built-in support for decoding JSON
responses, which can make API testing a bit cumbersome. The test clients
are therefore patched to add a JSON property:

.. code-block:: python

    def test_api(base_client):
        res = base_client.get(...)
        assert res.json == { ... }

Database re-use
---------------
The default database is an SQLite database located in the application's
instance folder. This can easily be overwritten by setting the environment
variable ``SQLALCHEMY_DATABASE_URI`` (useful e.g. in CI systems to test
multiple databases).

Tests that make changes to the database should explicitly use the function
scoped :py:data:`~fixtures.db` fixture. This fixture wraps the changes in
a transaction and rollback any changes by the end of the test. For instance:

.. code-block:: python

    def test_db1(db):
        db.session.add(User(username='alice'))
        db.session.commit()
        assert User.query.count() == 1 # i.e. independent of test_db2

    def test_db2(db):
        db.session.add(User(username='bob'))
        db.session.commit()
        assert User.query.count() == 1  # i.e. independent of test_db1


.. note::

    Take care! The :py:data:`~fixtures.db` fixture does not rollback other
    changes. If data, in addition to being added to the database, is also
    indexed in Elasticsearch then you should clear the index explicitly using
    e.g. :py:data:`~fixtures.es_clear`.

Performance considerations
~~~~~~~~~~~~~~~~~~~~~~~~~~
The database is recreated (all tables dropped and recreated) for each test
file, because the database is a module scoped fixture. This adds a performance
overhead, thus be careful not to indirectly depend on the database fixtures in
a file unless it is really necessary (e.g. via the :py:data:`~fixtures.app`
fixture).

Elasticsearch testing
---------------------
Pytest-Invenio depends on Invenio-Search and any mappings registered on
Invenio-Search will be created if you depend on the :py:data:`~fixtures.es`
fixture. The fixture is module scoped, meaning that any fixture you write to
e.g. load test data should likely also be module scoped.

Clearing changes
~~~~~~~~~~~~~~~~
Unlike the database fixture, which automatically rollback changes, you must
explicitly depend on the :py:data:`~fixtures.es_clear` fixture if you makes
changes to the indexes. This ensures that you leave the indexes in a clean
state for the next test. The :py:data:`~fixtures.es_clear` fixture will however
delete and recreate the indexes, and thus comes with a performance penalty if
used.

.. code-block:: python

    def test_es1(es_clear):
        # ...

Performance considerations
~~~~~~~~~~~~~~~~~~~~~~~~~~
As for the database fixtures, Elasticsearch indexes are deleted and recreated
for each test file (due to module scoped fixture). Thus be careful not to
indirectly depend on the database fixtures in a file unless it is really
necessary (e.g. via the :py:data:`~fixtures.app` fixture).

CLI testing
-----------
Pytest-Invenio provides two quick short cuts for easier testing Click-based
commands that require an application context (i.e. most commands).

The shortest version is to use the :py:data:`~fixtures.cli_runner` fixture:

.. code-block:: python

    def test_cmd(cli_runner):
        result = cli_runner(mycmd)
        assert result.exit_code == 0

The downside is that the Click CLIRunner is recreated for each call. This is
not necessary, so an alternative is to use the :py:data:`~fixtures.script_info`
fixture, which however is more verbose:

.. code-block:: python

    def test_cmd(script_info):
        runner = CliRunner()
        result = runner.invoke(mycmd, obj=script_info)
        assert result.exit_code == 0

Mail testing
------------
If you have Invenio-Mail installed on your application, you can use the
:py:data:`~fixtures.mailbox` fixture to test email sending. Any message sent
by the application during the test will be captured and is inspectable in via
the fixture:

.. code-block:: python

    def test_mailbox(appctx, mailbox):
        assert len(mailbox) == 0
        appctx.extensions['mail'].send_message(
            sender='no-reply@localhost',
            subject='testing',
            body='test',
            recipients=['no-reply@localhost'],)
        assert len(mailbox) == 1


End-to-end testing
------------------
In addition to using the Flask test client for testing views (see
:ref:`views-testing`), you can use a real browser via the Selenium integration
for fully end-to-end testing. The tests works by starting the Flask application
in a separate process, and using Selenium to drive your favorite browser.
Writing the tests are very easy, simply depend on the ``live_server`` fixture
(defined by pytest-flask) and the :py:data:`~fixtures.browser` fixture:

.. code-block:: python

    def test_browser(live_server, browser):
        # Note the use of '_external=True'
        browser.get(url_for('index', _external=True))

Running E2E tests
~~~~~~~~~~~~~~~~~
By default, tests using the :py:data:`~fixtures.browser` fixture are skipped.
In order to run these tests, you must set an environment variable:

.. code-block:: console

    $ export E2E=yes

Also, by default Chrome is used. If you'd like to use Firefox, Safari or
another browser you must set another environment variable:

.. code-block:: console

    $ export E2E_WEBDRIVER_BROWSERS="Firefox"

.. note::

    You must have Selenium Client and the Chrome Webdriver installed on your
    system in order to run the E2E tests.


Screenshots
~~~~~~~~~~~
The :py:data:`~fixtures.browser` fixture will take a screenshot of in case of
test failures and store it in a folder ``.e2e_screenshots``. On CI systems you
can also have screenshot printed to the console by setting an environment
variable:

.. code-block:: console

    $ export E2E_OUTPUT=base64

TravisCI integration
~~~~~~~~~~~~~~~~~~~~
Following is an example of the needed changes (at time of writing) to your
``.travis.yml`` in case want to run E2E tests on Travis. Travis is likely
to evolve, so please refer to the Travis CI documentation for the latest
information.

.. code-block:: yaml

    # Install Chrome
    # - see https://docs.travis-ci.com/user/chrome
    addons:
      chrome: stable

    # Chrome driver fails if not trusty dist
    dist: trusty

    # Selenium webdriver for Chrome fails if not on sudo
    # - see https://github.com/travis-ci/travis-ci/issues/8836
    sudo: true

    # Define environment variables to enable E2E tests and outputing
    # screenshots to the console.
    env:
      global:
        # Print screenshots to console output
        - E2E_OUTPUT=base64
        # Enable end-to-end tests
        - E2E=yes

    # Install Chrome webdriver for Selenium
    before_install:
      - "PATH=$PATH:$HOME/webdrivers"
      - "if [ ! -f $HOME/webdrivers/chromedriver ]; then wget https://chromedriver.storage.googleapis.com/2.31/chromedriver_linux64.zip -P   $HOME/webdrivers; unzip -d $HOME/webdrivers $HOME/webdrivers/chromedriver_linux64.zip; fi"  # noqa

    # Start a virtual display
    # - https://docs.travis-ci.com/user/gui-and-headless-browsers/
    before_script:
      - "export DISPLAY=:99.0"
      - "sh -e /etc/init.d/xvfb start"
      - sleep 3 # give xvfb some time to start
"""

from __future__ import absolute_import, print_function

from .version import __version__

__all__ = ('__version__', )



# -*- coding: utf-8 -*-
#
# This file is part of pytest-invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2018 Northwestern University, Feinberg School of Medicine,
# Galter Health Sciences Library.
#
# pytest-invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest plugin for Invenio.

The plugin adds fixtures to help with creation of applications as well as
configuring and initializing the database and search engine.

Additional the plugin helps with configuring end-to-end tests with selenium
and taking screenshots of failed selenium tests (useful for inspecting why
the test failed on CI systems).
"""

from __future__ import absolute_import, print_function

import os

import pytest

from .fixtures import _monkeypatch_response_class, app, app_config, appctx, \
    base_app, base_client, broker_uri, browser, bucket_from_dir, \
    celery_config_ext, cli_runner, database, db, db_uri, default_handler, es, \
    es_clear, instance_path, location, mailbox, script_info


def pytest_generate_tests(metafunc):
    """Skip end-to-end tests unless requested via ``E2E`` env variable.

    A screenshot is taken in case of test failures. Set the environment
    variable ``E2E_OUTPUT`` to ``base64`` to have the base64 encoded screenshot
    printed to stdout (useful in e.g. CI systems). Screenshots are saved to
    an ``.e2e_screenshots`` folder.

    Overrides pytest's default test collection function to skip tests using the
    ``browser`` fixture, unless the environment variable ``E2E`` is set to
    ``yes``.

    Each test using the ``browser`` fixture is parameterized with the list of
    browsers declared by the ``E2E_WEBDRIVER_BROWSERS`` environment variable.
    By default only Chrome is tested.
    """
    if 'browser' in metafunc.fixturenames:
        if os.environ.get('E2E', 'no').lower() != 'yes':
            pytest.skip(
                "End-to-end tests skipped because E2E environment variable "
                "was not set to 'yes'.")

        # Parameterize test based on list of browsers.
        browsers = os.environ.get('E2E_WEBDRIVER_BROWSERS', 'Chrome').split()
        metafunc.parametrize(
            'browser', browsers, indirect=True, scope='function'
        )


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Add hook to track if the test passed or failed."""
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    # used to e.g. during browser tests to take screenshot on failure
    setattr(item, "rep_" + rep.when, rep)



# -*- coding: utf-8 -*-
#
# This file is part of pytest-invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2018 Esteban J. G. Garbancho.
#
# pytest-invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest fixtures for Invenio."""

from __future__ import absolute_import, print_function

import json
import os
import shutil
import sys
import tempfile
from datetime import datetime

import pkg_resources
import pytest
from pytest_flask.plugin import _make_test_response_class
from selenium import webdriver

SCREENSHOT_SCRIPT = """import base64
with open('screenshot.png', 'wb') as fp:
    fp.write(base64.b64decode('''{data}'''))
"""


@pytest.fixture(scope='module')
def default_handler():
    """Flask default logging handler.

    Flask 0.13/1.0 changed logging to not add the default handler in case a
    handler is already installed. pytest automatically adds a handler to the
    root logger, causing Flask not to add a handler. This is an issue when
    testing Click output which uses the logger to output to the console.
    """
    try:
        from flask.logging import default_handler as handler
        return handler
    except ImportError:
        return None


@pytest.fixture(scope='module')
def instance_path():
    """Temporary instance path.

    Scope: module

    This fixture creates a temporary directory and sets the ``INSTANCE_PATH``
    environment variable to this directory. The directory is automatically
    removed.
    """
    path = tempfile.mkdtemp()
    os.environ.update(
        INVENIO_INSTANCE_PATH=os.environ.get('INSTANCE_PATH', path),
        INVENIO_STATIC_FOLDER=os.path.join(sys.prefix, 'var/instance/static'),
    )
    yield path
    os.environ.pop('INVENIO_INSTANCE_PATH', None)
    shutil.rmtree(path)


@pytest.fixture(scope='module')
def db_uri(instance_path):
    """Database URI (defaults to an SQLite datbase in the instance path).

    Scope: module

    The database can be overwritten by setting the ``SQLALCHEMY_DATABASE_URI``
    environment variable to a SQLAlchemy database URI.
    """
    if 'SQLALCHEMY_DATABASE_URI' in os.environ:
        yield os.environ['SQLALCHEMY_DATABASE_URI']
    else:
        filepath = tempfile.mkstemp(
            dir=instance_path, prefix='test', suffix='.db')[1]
        yield 'sqlite:///{}'.format(filepath)
        os.remove(filepath)


@pytest.fixture(scope='module')
def broker_uri():
    """Broker URI (defaults to an RabbitMQ on localhost).

    Scope: module

    The broker can be overwritten by setting the ``BROKER_URL`` environment
    variable.
    """
    yield os.environ.get('BROKER_URL', 'amqp://guest:guest@localhost:5672//')


def _celery_config():
    """Factory/Helper function to create the ``celery_config`` fixture.

    When Celery is installed it provides with this same fixture via
    `celery.contrib.pytest
    <https://github.com/celery/celery/blob/master/celery/contrib/pytest.py>`_,
    in this is the case we overwrite this fixture and update the configuration
    with Invenio's default configuration.
    If it is not installed, then we just define a new fixture which returns the
    default Invenio Celery configuration.
    """
    default_config = dict(
        CELERY_TASK_ALWAYS_EAGER=True,
        CELERY_CACHE_BACKEND='memory',
        CELERY_TASK_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_RESULT_BACKEND='cache',
    )

    try:
        pkg_resources.get_distribution('celery')

        # Celery is installed, overwrite fixture
        def inner(celery_config):
            celery_config.update(default_config)
            return celery_config
    except pkg_resources.DistributionNotFound:
        # No Celery, return the default config
        def inner():
            return default_config

    return inner


celery_config_ext = pytest.fixture(
    scope='module', name='celery_config_ext')(_celery_config())
"""Celery configuration (defaults to eager tasks).

Scope: module

This fixture provides the default Celery configuration (eager tasks,
in-memory result backend and exception propagation). It can easily be
overwritten in a specific test module:

.. code-block:: python

    # test_something.py
    import pytest

    pytest.fixture(scope='module')
    def celery_config_ext(celery_config_ext):
        celery_config_ext['CELERY_TASK_ALWAYS_EAGER'] = False
        return celery_config_ext
"""


@pytest.fixture(scope='module')
def app_config(db_uri, broker_uri, celery_config_ext):
    """Application configuration fixture.

    Scope: module

    This fixture sets default configuration for an Invenio application to make
    it suitable for testing. The database and broker URL are injected into the
    config, CSRF-protection in forms disabled, HTTP secure headers is disabled,
    mail sending is output to console.

    The fixture can easily be customized in your ``conftest.py`` or specific
    test module:

    .. code-block:: python

        # conftest.py
        import pytest

        pytest.fixture(scope='module')
        def app_config(app_config):
            app_config['MYVAR'] = 'test'
            return app_config
    """
    return dict(
        APP_DEFAULT_SECURE_HEADERS=dict(
            force_https=False,
            content_security_policy={'default-src': []}
        ),
        # Broker configuration
        BROKER_URL=broker_uri,
        # Disable Flask-DebugToolbar if installed.
        DEBUG_TB_ENABLED=False,
        # Disable mail sending.
        MAIL_SUPPRESS_SEND=True,
        # Allow testing OAuth without SSL.
        OAUTHLIB_INSECURE_TRANSPORT=True,
        OAUTH2_CACHE_TYPE='simple',
        # Set test secret keys
        SECRET_KEY='test-secret-key',
        SECURITY_PASSWORD_SALT='test-secret-key',
        # Database configuration
        SQLALCHEMY_DATABASE_URI=db_uri,
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        # Flask testing mode
        TESTING=True,
        # Disable CRSF protection in WTForms
        WTF_CSRF_ENABLED=False,
        # Celery configuration
        **celery_config_ext
    )


@pytest.fixture(scope='module')
def base_app(create_app, app_config, request, default_handler):
    """Base application fixture (without database, search and cache).

    Scope: module.

    This fixture is responsible for creating the Invenio application. It
    depends on an application factory fixture that must be defined by the user.

    .. code-block:: python

        # confest.py
        import pytest

        @pytest.fixture(scope='module)
        def create_app():
            from invenio_app.factory import create_api
            return create_api

    It is possible to overide the application factory for a specific test
    module, either by defining a fixture like above example, or simply setting
    the ``create_app`` property on the module:

    .. code-block:: python

        # test_something.py

        from invenio_app.factory import create_api
        create_app = create_api

        def test_acase(base_app):
            # ...
    """
    # Use create_app from the module if defined, otherwise use default
    # create_app fixture.
    create_app = getattr(request.module, 'create_app', create_app)
    app_ = create_app(**app_config)
    # See documentation for default_handler
    if default_handler:
        app_.logger.addHandler(default_handler)
    yield app_


@pytest.fixture(autouse=True, scope='function')
def _monkeypatch_response_class(request, monkeypatch):
    """Set custom response class to easily test JSON responses.

    .. code-block:: python

        def test_json(client):
            res = client.get(...)
            assert res.json == {'ping': 'pong'}

    Pytest-Flask provides this already for the "app" fixture
    """
    if 'base_app' not in request.fixturenames:
        return

    base_app = request.getfixturevalue('base_app')
    monkeypatch.setattr(
        base_app, 'response_class',
        _make_test_response_class(base_app.response_class))


@pytest.fixture(scope='function')
def base_client(base_app):
    """Test client for the base application fixture.

    Scope: function

    If you need the database and search indexes initialized, simply use the
    Pytest-Flask fixture ``client`` instead. This fixture is mainly useful if
    you need a test client without needing to initialize both the database and
    search indexes.
    """
    with base_app.test_client() as client:
        yield client


@pytest.fixture(scope='module')
def appctx(base_app):
    """Application context for the current base application.

    Scope: module

    This fixture pushes an application context on the stack, so that
    ``current_app`` is defined and e.g ``url_for`` will also work.
    """
    with base_app.app_context():
        yield base_app


@pytest.fixture(scope='module')
def script_info(base_app):
    """Get ScriptInfo object for testing a CLI command.

    Scope: module

    .. code-block:: python

        def test_cmd(script_info):
            runner = CliRunner()
            result = runner.invoke(mycmd, obj=script_info)
            assert result.exit_code == 0
    """
    from flask.cli import ScriptInfo
    return ScriptInfo(create_app=lambda info: base_app)


@pytest.fixture(scope='module')
def cli_runner(script_info):
    """Create a CLI runner for testing a CLI command.

    Scope: module

    .. code-block:: python

        def test_cmd(cli_runner):
            result = cli_runner(mycmd)
            assert result.exit_code == 0
    """
    from click.testing import CliRunner

    def cli_invoke(command, input=None, *args):
        return CliRunner().invoke(command, args, input=input, obj=script_info)
    return cli_invoke


def _es_create_indexes(current_search, current_search_client):
    """Create all registered Elasticsearch indexes."""
    from elasticsearch.exceptions import RequestError
    try:
        list(current_search.create())
    except RequestError:
        list(current_search.delete(ignore=[404]))
        list(current_search.create())
    current_search_client.indices.refresh()


def _es_delete_indexes(current_search):
    """Delete all registered Elasticsearch indexes."""
    list(current_search.delete(ignore=[404]))


@pytest.fixture(scope='module')
def es(appctx):
    """Setup and teardown all registered Elasticsearch indices.

    Scope: module

    This fixture will create all registered indexes in Elasticsearch and remove
    once done. Fixtures that perform changes (e.g. index or remove documents),
    should used the function-scoped :py:data:`es_clear` fixture to leave the
    indexes clean for the following tests.
    """
    from invenio_search import current_search, current_search_client
    _es_create_indexes(current_search, current_search_client)
    yield current_search_client
    _es_delete_indexes(current_search)


@pytest.fixture(scope='function')
def es_clear(es):
    """Clear Elasticsearch indices after test finishes (function scope).

    Scope: function

    This fixture rollback any changes performed to the indexes during a test,
    in order to leave Elasticsearch in a clean state for the next test.
    """
    from invenio_search import current_search, current_search_client
    yield es
    _es_delete_indexes(current_search)
    _es_create_indexes(current_search, current_search_client)


@pytest.fixture(scope='module')
def database(appctx):
    """Setup database.

    Scope: module

    Normally, tests should use the function-scoped :py:data:`db` fixture
    instead. This fixture takes care of creating the database/tables and
    removing the tables once tests are done.
    """
    from invenio_db import db as db_
    from sqlalchemy_utils.functions import create_database, database_exists
    if not database_exists(str(db_.engine.url)):
        create_database(str(db_.engine.url))
    db_.create_all()

    yield db_

    db_.session.remove()
    db_.drop_all()


@pytest.fixture(scope='function')
def db(database):
    """Creates a new database session for a test.

    Scope: function

    You must use this fixture if your test connects to the database. The
    fixture will set a save point and rollback all changes performed during
    the test (this is much faster than recreating the entire database).
    """
    import sqlalchemy as sa
    connection = database.engine.connect()
    transaction = connection.begin()

    options = dict(bind=connection, binds={})
    session = database.create_scoped_session(options=options)

    session.begin_nested()

    # `session` is actually a scoped_session. For the `after_transaction_end`
    # event, we need a session instance to listen for, hence the `session()`
    # call.
    @sa.event.listens_for(session(), 'after_transaction_end')
    def restart_savepoint(sess, trans):
        if trans.nested and not trans._parent.nested:
            session.expire_all()
            session.begin_nested()

    old_session = database.session
    database.session = session

    yield database

    session.remove()
    transaction.rollback()
    connection.close()
    database.session = old_session


@pytest.fixture(scope='function')
def mailbox(base_app):
    """Mailbox fixture.

    Scope: function

    This fixture provides a mailbox that captures all outgoing emails and thus
    easily allow you to test mail sending in your app:

    .. code-block:: python

        def test_mailbox(appctx, mailbox):
            appctx.extensions['mail'].send_message(
                sender='no-reply@localhost',
                subject='Testing',
                body='Test',
                recipients=['no-reply@localhost'])
            assert len(mailbox) == 1
    """
    ext = base_app.extensions.get('mail')
    if ext is None:
        raise RuntimeError(
            'Invenio-Mail extension is not installed on application.')
    else:
        with ext.record_messages() as outbox:
            yield outbox


@pytest.fixture(scope='module')
def app(base_app, es, database):
    """Invenio application with database and Elasticsearch.

    Scope: module

    See also :py:data:`base_app` for an Invenio application fixture that
    does not initialize database and Elasticsearch.
    """
    yield base_app


@pytest.fixture(scope='session')
def browser(request):
    """Selenium webdriver fixture.

    Scope: session

    The fixture initializes a Selenium webdriver which can be used for
    end-to-end testing of your application:

    .. code-block:: python

        from flask import url_for

        def test_browser(live_server, browser):
            browser.get(url_for('index', _external=True))

    The ``live_server`` fixture is provided by Pytest-Flask and uses the
    :py:data:`app` fixture to determine which application to start.

    .. note::

        End-to-end test are only executed if the environment variable ``E2E``
        is set to yes::

            $ export E2E=yes

        This allows you to easily switch on/off end-to-end tests.

    By default, a Chrome webdriver client will be created. However, you can
    customize which browsers to test via the ``E2E_WEBDRIVER_BROWSERS``
    environment variable:

    .. code-block:: console

        $ export E2E_WEBDRIVER_BROWSERS="Chrome Firefox"

    If multiple browsers are requested, each test case using the
    :py:data:`browser` fixture will be parameterized with the list of browsers.

    In case the test fail, a screenshot will be taken and saved in folder
    ``.e2e_screenshots``.
    """
    browser_name = getattr(request, 'param', 'Chrome')
    driver = getattr(webdriver, browser_name)()

    yield driver

    _take_screenshot_if_test_failed(driver, request)

    # Quit the webdriver instance
    driver.quit()


def _take_screenshot_if_test_failed(driver, request):
    """Take a screenshot if the test failed."""
    if request.node.rep_call.failed:
        filename = '{modname}::{funname}::{now}.png'.format(
            modname=request.module.__name__,
            funname=request.function.__name__ if request.function else '',
            now=datetime.now().isoformat())
        filepath = os.path.join(_get_screenshots_dir(), filename)
        driver.get_screenshot_as_file(filepath)
        print("Screenshot of failing test:")
        if os.environ.get('E2E_OUTPUT') == 'base64':
            print(SCREENSHOT_SCRIPT.format(
                data=driver.get_screenshot_as_base64()))
        else:
            print(filepath)


def _get_screenshots_dir():
    """Create the screenshots directory."""
    directory = ".e2e_screenshots"
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


@pytest.yield_fixture(scope='function')
def location(db):
    """Creates a simple default location for a test.

    Scope: function

    Use this fixture if your test requires a `files location <https://invenio-
    files-rest.readthedocs.io/en/latest/api.html#invenio_files_rest.models.
    Location>`_. The location will be a default location with the name
    ``pytest-location``.
    """
    from invenio_files_rest.models import Location
    uri = tempfile.mkdtemp()
    location_obj = Location(name="pytest-location", uri=uri, default=True)

    db.session.add(location_obj)
    db.session.commit()

    yield location_obj

    shutil.rmtree(location_obj.uri)


@pytest.fixture(scope="function")
def bucket_from_dir(db, location):
    '''Creates a bucket from the specified directory.

    Scope: function

    Use this fixture if your test requires a `files bucket <https://invenio-
    files-rest.readthedocs.io/en/latest/api.html#invenio_files_rest.models.
    Bucket>`_. The ``bucket_from_dir`` fixture returns a function with the
    following signature:

    .. code-block:: python

        def create_bucket_from_dir(source_dir, location_obj=None):
            """Create bucket from the specified source directory.

            :param source_dir: The directory to create the bucket from.
            :param location_obj: Optional location object to use. If None
                is specified, get the current default location.
            :returns: The new bucket object.
            """

    Below is an example of how to use the ``bucket_from_dir`` fixture:

    .. code-block:: python

        def test_with_bucket(bucket_from_dir):
            bucket = bucket_from_dir('/my/directory/path')
            # ... use the bucket for your test
    '''
    def create_bucket_from_dir(source_dir, location_obj=None):
        """Create bucket from the specified source directory.

        :param source_dir: The directory to create the bucket from.
        :param location_obj: Optional location object to use. If None
            is specified, get the current default location.
        :returns: The new bucket object.
        """
        if not location_obj:
            from invenio_files_rest.models import Bucket, Location, \
                ObjectVersion
            location_obj = Location.get_default() or location
        bucket_obj = Bucket.create(location_obj)
        for file_name in os.listdir(source_dir):
            full_file_path = os.path.join(source_dir, file_name)
            if os.path.isdir(full_file_path):
                continue
            file_obj = open(full_file_path, 'rb')
            ObjectVersion.create(bucket_obj, key=file_name, stream=file_obj)
        db.session.commit()
        return bucket_obj
    return create_bucket_from_dir
